import 'dart:convert';

import 'package:flutter_inappwebview/flutter_inappwebview.dart';

import 'appier_flutter_platform.dart';

class WebSdkMessage {
  WebSdkMessage.fromJson(Map<String, dynamic> json)
      : name = json['name'] as String,
        body = json['body'];
  final String name;
  final dynamic body;
}

class WebSdkRequest {
  WebSdkRequest.fromJson(Map<String, dynamic> json)
      : id = json['id'] as String,
        name = json['name'] as String,
        body = json['body'];
  final String id;
  final String name;
  final dynamic body;
}

abstract class WebSdkBridge {
  static const String _script = """
if (!window.aiqMobileSdk) {
  console.log('[WebSdkBridge] Injecting...');
  window.aiqMobileSdk = {
    log: function(message, ...args) {
      console.debug('[WebSdkBridge] ' + message, ...args);
    },
    getPlatform: function() {
      return Promise.resolve('flutter');
    },
    getVersion: function() {
      return this.Promise.request('getVersion');
    },
    logEvent: function(jsonString) {
      this.Promise.request('logEvent', JSON.parse(jsonString));
    },
    setCustomKey: function(jsonString) {
      this.Promise.request('setCustomKey', JSON.parse(jsonString));
    },
    getRecommendationByScenario: function(jsonString) {
      return this.Promise.request('getRecommendationByScenario', JSON.parse(jsonString));
    },
    Message: {
      queue: [],
      log: function(...args) {
        window.aiqMobileSdk.log(...args);
      },
      post: function(name, body) {
        this.queue.push({name: name, body: body});
        this.flush();
      },
      flush: function() {
        if (!window.flutter_inappwebview || !window.flutter_inappwebview.callHandler) {
          return;
        }
        while (this.queue.length > 0) {
          var message = this.queue.shift();
          this.log('Posting message:', JSON.stringify(message));
          window.flutter_inappwebview.callHandler('aiqua', message);
        }
      },
      receive: function(name, jsonBody) {
        this.log('Receiving message:', name, jsonBody);
        var body = JSON.parse(jsonBody);
        switch (name) {
          case 'resolve':
            window.aiqMobileSdk.Promise.resolve(body.id, body.data);
            break;
          case 'reject':
            window.aiqMobileSdk.Promise.reject(body.id, body.error);
            break;
          default:
            this.log('Unexpected message:', name, body);
            break;
        }
      },
    },
    Promise: {
      queue: {},
      log: function(...args) {
        window.aiqMobileSdk.log(...args);
      },
      request: function(name, body = null, timeout = 15000) {
        var id = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(s) {
          var r = (Math.floor(new Date().getTime() / 16) + 16 * Math.random()) % 16 | 0;
          return ('x' == s ? r : 3 & r | 8).toString(16);
        });
        var self = this;
        setTimeout(function() {
          self.reject(id, 'Timed out after ' + timeout + 'ms', false);
        }, timeout);
        return new Promise(function(resolve, reject) {
          self.queue[id] = {
            name: name,
            body: body,
            resolve: resolve,
            reject: reject
          };
          window.aiqMobileSdk.Message.post('request', {id: id, name: name, body: body});
        });
      },
      resolve: function(id, data) {
        var promise = this.queue[id];
        if (!promise) {
          this.log('No such promise to resolve:', id);
          return;
        }
        this.log('Resolving promise:', id, promise.name, JSON.stringify(promise.body), JSON.stringify(data));
        promise.resolve(JSON.stringify(data));
        delete this.queue[id];
      },
      reject: function(id, error, important = true) {
        var promise = this.queue[id];
        if (!promise) {
          if (important) this.log('No such promise to reject: ' + id);
          return;
        }
        this.log('Rejecting promise:', id, promise.name, JSON.stringify(promise.body), error);
        promise.reject(Error(error));
        delete this.queue[id];
      },
    },
  };
  window.addEventListener("flutterInAppWebViewPlatformReady", function(event) {
    window.aiqMobileSdk.Message.flush();
  });
  console.log('[WebSdkBridge] Injected.');
}
""";

  void _post(String name, dynamic body);

  void _respond(WebSdkRequest request) {
    switch (request.name) {
      case 'getVersion':
        getVersion()
            .then((version) => _resolve(request.id, version))
            // ignore: avoid_types_on_closure_parameters
            .catchError((Object e) => _reject(request.id, e));
        break;
      case 'logEvent':
        final body = request.body as Map<String, dynamic>;
        final name = body['eventName'] as String;
        final parameters = body['parameters'] as Map<String, dynamic>?;
        final dynamic vtsValue = body['vts'];
        double? vts;
        if (vtsValue is int) {
          vts = vtsValue.toDouble();
        } else if (vtsValue is double) {
          vts = vtsValue;
        }
        final vtsCurr = body['vtsCurr'] as String?;
        AppierFlutterPlatform.instance
            .logEvent(name, parameters: parameters, vts: vts, vtsCurr: vtsCurr)
            .then((_) => _resolve(request.id, null))
            // ignore: avoid_types_on_closure_parameters
            .catchError((Object e) => _reject(request.id, e));
        break;
      case 'setCustomKey':
        final body = request.body as Map<String, dynamic>;
        final key = body['key'] as String;
        final dynamic value = body['value'];
        AppierFlutterPlatform.instance
            .setCustomKey(key, value)
            .then((_) => _resolve(request.id, null))
            // ignore: avoid_types_on_closure_parameters
            .catchError((Object e) => _reject(request.id, e));
        break;
      case 'getRecommendationByScenario':
        final body = request.body as Map<String, dynamic>;
        final scenarioId = body['scenarioId'] as String;
        final productId = body['productId'] as String?;
        final filter = body['filter'] as Map<String, dynamic>?;
        // it can be null, string, or int. Please check https://appier.atlassian.net/browse/QGWL-14101
        final dynamic num = body['num'];
        final clientUserId = body['user_id'] as String?;
        final parameters = <String, dynamic>{};

        if (filter != null) {
          parameters['filter'] = filter;
        }
        if (num is String && int.parse(num) > 0) {
          parameters['num'] = int.parse(num);
        } else if (num is int && num > 0) {
          parameters['num'] = num;
        }
        if (clientUserId != null && clientUserId != '') {
          parameters['user_id'] = clientUserId;
        }

        AppierFlutterPlatform.instance
            .getRecommendation(scenarioId,
                productId: productId, parameters: parameters)
            .then((data) => _resolve(request.id, data))
            // ignore: avoid_types_on_closure_parameters
            .catchError((Object e) => _reject(request.id, e));
        break;
      default:
        throw Exception('Unexpected request: ${request.name}');
    }
  }

  Future<String> getVersion() {
    return Future.value('1.0.0');
  }

  void _resolve(String id, dynamic data) {
    _post('resolve', <String, dynamic>{'id': id, 'data': data});
  }

  void _reject(String id, Object error) {
    _post('reject', {'id': id, 'error': '$error'});
  }
}

class InAppWebViewBridge extends WebSdkBridge {
  InAppWebViewBridge(this._controller);
  final InAppWebViewController _controller;
  static final UserScript _userScript = UserScript(
      source: WebSdkBridge._script,
      injectionTime: UserScriptInjectionTime.AT_DOCUMENT_START);
  Future<void> install() async {
    await _controller.removeUserScript(userScript: _userScript);
    await _controller.addUserScript(userScript: _userScript);
    await _controller.evaluateJavascript(source: _userScript.source);
    _controller.javaScriptHandlersMap.remove('aiqua');
    _controller.addJavaScriptHandler(
        handlerName: 'aiqua',
        callback: (arguments) {
          if (arguments.isEmpty) {
            print('No argument to handle.');
            return null;
          }
          print('Handling message: ${arguments.first}');
          final message =
              WebSdkMessage.fromJson(arguments.first as Map<String, dynamic>);
          final name = message.name;
          final dynamic body = message.body;
          switch (name) {
            case 'request':
              final request =
                  WebSdkRequest.fromJson(body as Map<String, dynamic>);
              try {
                _respond(request);
                // ignore: avoid_catches_without_on_clauses
              } catch (e) {
                _reject(request.id, e);
              }
              break;
            default:
              print('Unexpected message: $name');
              break;
          }
        });
  }

  @override
  void _post(String name, dynamic body) {
    final jsonBody = jsonEncode(body);
    print('Posting message: $name => $jsonBody');
    _controller.evaluateJavascript(
        source: "window.aiqMobileSdk.Message.receive('$name', `$jsonBody`)");
  }
}
